@@ -24,6 +24,11 @@ from scipy.maxentropy import logsumexp # log of sum
 from gensim import interfaces, utils
 
 
+logger = logging.getLogger('ldamodel')
+logger.setLevel(logging.INFO)
+
+
+
 trigamma = lambda x: polygamma(1, x) # second derivative of the gamma fnc
 
 
@@ -63,18 +68,17 @@ class LdaModel(interfaces.TransformationABC):
         """
         # store user-supplied parameters
         self.id2word = id2word
+
         if self.id2word is None:
-            logging.info("no word id mapping provided; initializing from corpus, assuming identity")
-            maxId = -1
-            for document in corpus:
-                maxId = max(maxId, max([-1] + [fieldId for fieldId, _ in document]))
-            self.numTerms = 1 + maxId
-            self.id2word = dict(zip(xrange(self.numTerms), xrange(self.numTerms)))
+            logger.info("no word id mapping provided; initializing from corpus, assuming identity")
+            self.id2word = utils.dictFromCorpus(corpus)
+            self.numTerms = len(self.id2word)
         else:
             self.numTerms = 1 + max([-1] + self.id2word.keys())
+
         self.numTopics = numTopics # number of latent topics
         
-        # internal constants; can be manually changed after having called the init
+        # internal constants; can be manually changed after having called this constructor and before calling `initialize()`
         self.ESTIMATE_ALPHA = alpha is None
         if alpha is None: # no alpha supplied by user => get some initial estimate
             alpha = 10.0 / numTopics # initial estimate is 50 / numTopics, as suggested in Steyvers&Griffiths: Probabilistic Topic Models
@@ -108,7 +112,7 @@ class LdaModel(interfaces.TransformationABC):
         corpus, and is thus much slower.
         """
         # initialize the model
-        logging.info("initializing LDA model with '%s'" % initMode)
+        logger.info("initializing LDA model with '%s'" % initMode)
 
         # set initial word counts
         if self.numTerms == 0:
@@ -127,11 +131,11 @@ class LdaModel(interfaces.TransformationABC):
         
         # main EM loop
         for i in xrange(self.EM_MAX_ITER):
-            logging.info("starting EM iteration #%i, converged=%s, likelihood=%s" % 
+            logger.info("starting EM iteration #%i, converged=%s, likelihood=%s" % 
                          (i, converged, likelihood))
 
             if numpy.isfinite(converged) and (converged <= self.EM_CONVERGED): # solution good enough?
-                logging.info("EM converged in %i iterations" % i)
+                logger.info("EM converged in %i iterations" % i)
                 break
             
             # initialize help structures for this iteration
@@ -140,17 +144,17 @@ class LdaModel(interfaces.TransformationABC):
             self.numDocs = 0
             
             # E step: iterate over individual documents, using old beta and updating new word counts
-            logging.info("performing E step #%i" % i)
+            logger.info("performing E step #%i" % i)
             likelihood = sum(self.docEStep(doc) for doc in corpus)
             assert numpy.isfinite(likelihood), "bad likelihood %s" % likelihood
 
             # M step -- update alpha and beta
-            logging.info("performing M step #%i" % i)
+            logger.info("performing M step #%i" % i)
             self.mle(estimateAlpha = self.ESTIMATE_ALPHA)
             
             # check for convergence
             converged = numpy.divide(likelihoodOld - likelihood, likelihoodOld)
-            logging.info("finished iteration #%i: likelihood %f, likelihoodOld %f, converged %f" % 
+            logger.info("finished iteration #%i: likelihood %f, likelihoodOld %f, converged %f" % 
                          (i, likelihood, likelihoodOld, converged))
             likelihoodOld = likelihood
 
@@ -186,7 +190,7 @@ class LdaModel(interfaces.TransformationABC):
         if estimateAlpha:
             self.alpha = self.optAlpha()
         
-        logging.debug("updated model to %s" % self)
+        logger.debug("updated model to %s" % self)
     
     
     def optAlpha(self, MAX_ALPHA_ITER = 1000, NEWTON_THRESH = 1e-5):
@@ -196,24 +200,24 @@ class LdaModel(interfaces.TransformationABC):
         """
         initA = 100.0
         logA = numpy.log(initA) # keep computations in log space
-        logging.debug("optimizing old alpha %s" % self.alpha)
+        logger.debug("optimizing old alpha %s" % self.alpha)
         
         for i in xrange(MAX_ALPHA_ITER):
             a = numpy.exp(logA)
             if not numpy.isfinite(a):
                 initA = initA * 10.0
-                logging.warning("alpha is NaN; new init alpha=%f" % initA)
+                logger.warning("alpha is NaN; new init alpha=%f" % initA)
                 a = initA
                 logA = numpy.log(a)
             f = self.numDocs * (gammaln(self.numTopics * a) - self.numTopics * gammaln(a)) + (a - 1) * self.alphaSuffStats
             df = self.alphaSuffStats + self.numDocs * (self.numTopics * digamma(self.numTopics * a) - self.numTopics * digamma(a))
             d2f = self.numDocs * (self.numTopics * self.numTopics * trigamma(self.numTopics * a) - self.numTopics * trigamma(a))
             logA -= df / (d2f * a + df)
-#            logging.debug("alpha maximization: f=%f, df=%f" % (f, df))
+#            logger.debug("alpha maximization: f=%f, df=%f" % (f, df))
             if numpy.abs(df) <= NEWTON_THRESH:
                 break
         result = numpy.exp(logA) # convert back from log space
-        logging.info("estimated old alpha %s to new alpha %s" % (self.alpha, result))
+        logger.info("estimated old alpha %s to new alpha %s" % (self.alpha, result))
         return result
 
 
@@ -238,11 +242,11 @@ class LdaModel(interfaces.TransformationABC):
         
         # variational estimate
         for i in xrange(self.VAR_MAX_ITER):
-#            logging.debug("inference step #%s, converged=%s, likelihood=%s, likelikelihoodOld=%s" % 
+#            logger.debug("inference step #%s, converged=%s, likelihood=%s, likelikelihoodOld=%s" % 
 #                          (i, converged, likelihood, likelihoodOld))
             
             if numpy.isfinite(converged) and converged <= self.VAR_CONVERGED:
-                logging.debug("document converged in %i iterations" % i)
+                logger.debug("document converged in %i iterations" % i)
                 break
             
             for n, (wordIndex, wordCount) in enumerate(doc):
@@ -288,12 +292,12 @@ class LdaModel(interfaces.TransformationABC):
         Initialize the model word counts from the corpus. Each topic will be initialized
         from `numInitDocs` random documents.
         """
-        logging.info("initializing model with %i random document(s) per topic" % numInitDocs)
+        logger.info("initializing model with %i random document(s) per topic" % numInitDocs)
         result = numpy.ones(shape = (self.numTopics, self.numTerms)) # add-one smooth
         # next we precompute the all the random document indices, so that we can 
         # update the counts in a single sweep over the corpus. 
         # all this drama is because the corpus doesn't necessarily support 
-        # random access (indexing) -- it only supports sequential iteration over 
+        # random access -- it only supports sequential iteration over 
         # the documents (streaming).
         initDocs = numpy.random.randint(0, len(corpus), (self.numTopics, numInitDocs)) # get document indices
         
@@ -320,7 +324,7 @@ class LdaModel(interfaces.TransformationABC):
         Topics are sorted by probability, words are in the same order as in the input.
         """
         fname = corpus.fname + '.lda_inferred'
-        logging.info("writing inferences to %s" % fname)
+        logger.info("writing inferences to %s" % fname)
         fout = open(fname, 'w')
         for doc in corpus:
             # do the inference
@@ -403,8 +407,9 @@ class LdaModel(interfaces.TransformationABC):
         Ignore topics with very low probability (below 0.001).
         """
         # if the input vector is in fact a corpus, return a transformed corpus as result
-        if utils.isCorpus(bow):
-            return self._apply(bow)
+        is_corpus, corpus = utils.isCorpus(bow)
+        if is_corpus:
+            return self._apply(corpus)
         
         likelihood, phi, gamma = self.inference(bow)
         gamma -= self.alpha # subtract topic prior, to get the expected number of words for each topic
