@@ -4,8 +4,45 @@
 # Copyright (C) 2010 Radim Rehurek <radimrehurek@seznam.cz>
 # Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html
 
+
 """
 Module for Latent Semantic Indexing.
+
+It actually contains several algorithms for decomposition of extremely large 
+corpora, a combination of which effectively and transparently allows building LSI
+models for:
+
+* corpora much larger than RAM: only constant memory needed, independent of 
+  the corpus size (though still dependent on the feature set size)
+* corpora that are streamed: documents are only accessed sequentially, not 
+  random-accessed
+* corpora that cannot even be temporarily stored, each document can only be 
+  seen once and must be processed immediately (one-pass algorithm)
+* distributed computing for ultra large corpora, making use of a cluster of
+  machines
+
+Wall-clock performance on the English Wikipedia (2G corpus positions, 3.2M 
+documents, 100K features, 0.5G non-zero entries in the final TF-IDF matrix), 
+requesting the top 400 LSI factors:
+
+
+======================================== ============ ==================
+ algorithm                                 serial      distributed    
+======================================== ============ ==================
+ one-pass update algo (chunks=factors)     109h        19h            
+ one-pass merge algo (chunks=40K docs)     8.5h        2.3h           
+ two-pass randomized algo (chunks=40K)     2.5h        N/A [1]_           
+======================================== ============ ==================
+
+*serial* = Core 2 Duo MacBook Pro 2.53Ghz, 4GB RAM, libVec
+
+*distributed* = cluster of six logical nodes on four physical machines, each with 
+dual core Xeon 2.0GHz, 4GB RAM, ATLAS
+
+.. [1] The two-pass algo could be distributed too, but most time is already spent 
+   reading/decompressing the input from disk, and the extra network traffic due to 
+   data distribution would likely make it actually *slower*.
+
 """
 
 
@@ -14,11 +51,11 @@ import itertools
 
 import numpy
 import scipy.sparse
-
-# scipy is still an experimental package and locations change, so try to work
-# around differences (currently only triu in scipy 0.7 vs. 0.8)
-from scipy.linalg.blas import get_blas_funcs
+from scipy.sparse import sparsetools
 from scipy.linalg.lapack import get_lapack_funcs, find_best_lapack_type
+
+# scipy is not a stable package yet, locations change, so try to work
+# around differences (currently only concerns location of 'triu' in scipy 0.7 vs. 0.8)
 try:
     from scipy.linalg.basic import triu
 except ImportError:
@@ -33,56 +70,69 @@ logger.setLevel(logging.INFO)
 
 
 
-def clipSpectrum(s, k):
+def clipSpectrum(s, k, discard = 0.001):
     """
-    Given singular values `s`, return how many factors should be kept to avoid
-    storing spurious values. The returned value is at most `k`.
+    Given eigenvalues `s`, return how many factors should be kept to avoid
+    storing spurious (tiny, numerically instable) values. 
+    
+    This will ignore the tail of the spectrum with relative combined mass < min(`discard`, 1/k).
+    
+    The returned value is clipped against `k` (= at most `k` factors).
     """
-    # compute relative contribution of each singular value towards the energy spectrum
-    rel_spectrum = numpy.abs(1.0 - numpy.cumsum(s**2 / numpy.sum(s ** 2)))
-    # ignore the last 0.1% (or 1/k percent, whichever is smaller) of the spectrum
-    small = 1 + len(numpy.where(rel_spectrum > min(0.001, 1.0 / k))[0]) 
+    # compute relative contribution of eigenvalues towards the energy spectrum
+    rel_spectrum = numpy.abs(1.0 - numpy.cumsum(s / numpy.sum(s)))
+    # ignore the last `discard` (or 1/k, whichever is smaller) of the spectrum
+    small = 1 + len(numpy.where(rel_spectrum > min(discard, 1.0 / k))[0]) 
     k = min(k, small) # clip against k
-    logger.info("keeping %i factors (discarding %.2f%% of energy spectrum)" %
+    logger.info("keeping %i factors (discarding %.3f%% of energy spectrum)" %
                 (k, 100 * rel_spectrum[k - 1]))
     return k
 
 
 class Projection(utils.SaveLoad):
-    def __init__(self, m, k, docs = None):
+    def __init__(self, m, k, docs = None, algo = 'onepass', chunks = None):
         """
-        Store (U, S) projection itself. This is the class taking care of 'core math';
-        interfacing with corpora, training etc is done through class LsiModel.
+        Construct the (U, S) projection from a corpus `docs`. 
         
-        `docs` is either a spare matrix or a corpus which, when converted to a 
-        sparse matrix, must fit comfortably into main memory.
+        This is the class taking care of 'core math'; interfacing with corpora, 
+        chunking large corpora etc. is done through the LsiModel class.
+        
+        `algo` is currently one of:
+        
+          * 'onepass'; only a single pass over `docs` is needed
+          * 'twopass'; multiple passes over the input allowed => can use a 
+            faster algorithm.
         """
-
         self.m, self.k = m, k
         if docs is not None:
-            # base case decomposition: given a job `docs`, compute its decomposition 
-            # in core, algorithm 1
-            if utils.isCorpus(docs):
-                docs = matutils.corpus2csc(m, docs)
-            if docs.shape[1] <= max(k, 100):
-                # for sufficiently small chunk size, compute svd(now, a) instead of svd(now, svd(a)).
-                # this improves accuracy and is also faster for small chunks, because
-                # we need to do one less svd.
-                # on larger chunks this doesn't work because we run out of memory (chunks=1000 
-                # would already raise MemoryException on my machine)
-                self.u = docs
-                self.s = None
+            # base case decomposition: given a job `docs`, compute its decomposition in-core
+            # results of several base case decompositions can be merged via `self.merge()`
+            if algo == 'twopass':
+                self.u, self.s = stochasticSvd(docs, k, chunks = chunks, num_terms = m)
+            elif algo == 'onepass':
+                if not scipy.sparse.issparse(docs):
+                    docs = matutils.corpus2csc(docs, num_terms = m)
+                if docs.shape[1] <= max(k, 100):
+                    # For sufficiently small chunk size, update directly like `svd(now, docs)` 
+                    # instead of `svd(now, svd(docs))`.
+                    # This improves accuracy and is also faster for small chunks, because
+                    # we need to perform one less svd.
+                    # On larger chunks this doesn't work because we quickly run out of memory.
+                    self.u = docs
+                    self.s = None
+                else:
+                    try:
+                        import sparsesvd
+                    except ImportError:
+                        raise ImportError("for LSA, the `sparsesvd` module is needed but not found; run `easy_install sparsesvd`")
+                    logger.info("computing sparse SVD of %s matrix" % str(docs.shape))
+                    ut, s, vt = sparsesvd.sparsesvd(docs, k + 30) # ask for extra factors, because for some reason SVDLIBC sometimes returns fewer factors than requested
+                    u = ut.T
+                    del ut, vt
+                    k = clipSpectrum(s ** 2, self.k)
+                    self.u, self.s = u[:, :k], s[:k]
             else:
-                try:
-                    import sparsesvd
-                except ImportError:
-                    raise ImportError("for LSA, the `sparsesvd` module is needed but not found; run `easy_install sparsesvd`")
-                logger.info("computing sparse SVD of %s matrix" % str(docs.shape))
-                ut, s, vt = sparsesvd.sparsesvd(docs, k + 30) # ask for extra factors, because for some reason SVDLIBC sometimes returns fewer factors than requested
-                u = ut.T
-                del ut, vt
-                k = clipSpectrum(s, self.k)
-                self.u, self.s = u[:, :k], s[:k]
+                raise NotImplementedError("unknown decomposition algorithm: '%s'" % algo)
         else:
             self.u, self.s = None, None
     
@@ -95,10 +145,8 @@ class Projection(utils.SaveLoad):
         """
         Merge this Projection with another. 
         
-        Content of `other` is destroyed in the process, so pass this function a 
-        copy if you need it further.
-        
-        This is the optimized merge described in algorithm 5.
+        The content of `other` is destroyed in the process, so pass this function a 
+        copy of `other` if you need it further.
         """
         if other.u is None:
             # the other projection is empty => do nothing
@@ -124,7 +172,7 @@ class Projection(utils.SaveLoad):
                     u = ut.T
                     del ut
                 del vt
-                k = clipSpectrum(s, self.k)
+                k = clipSpectrum(s ** 2, self.k)
                 self.u = u[:, :k].copy('F')
                 self.s = s[:k]
             else:
@@ -140,15 +188,15 @@ class Projection(utils.SaveLoad):
             other.u = other.u.todense()
             other.s = 1.0 # broadcasting will promote this to eye(n2) where needed
         # TODO Maybe keep the bases as elementary reflectors, without 
-        # forming explicit matrices with gorgqr.
+        # forming explicit matrices with ORGQR.
         # The only operation we ever need is basis^T*basis ond basis*component.
-        # But how to do that in numpy? And is it fast(er)?
+        # But how to do that in scipy? And is it fast(er)?
         
         # find component of u2 orthogonal to u1
-        # IMPORTANT: keep matrices in suitable order for matrix products; failing to do so gives 8x lower performance :(
+        # IMPORTANT: keep matrices in memory suitable order for matrix products; failing to do so gives 8x lower performance :(
         self.u = numpy.asfortranarray(self.u) # does nothing if input already fortran-order array
         other.u = numpy.asfortranarray(other.u)
-        gemm, = get_blas_funcs(('gemm',), (self.u,))
+        gemm = matutils.blas('gemm', self.u)
         logger.debug("constructing orthogonal component")
         c = gemm(1.0, self.u, other.u, trans_a = True)
         gemm(-1.0, self.u, c, beta = 1.0, c = other.u, overwrite_c = True)
@@ -156,8 +204,8 @@ class Projection(utils.SaveLoad):
         # perform q, r = QR(component); code hacked out of scipy.linalg.qr
         logger.debug("computing QR of %s dense matrix" % str(other.u.shape))
         geqrf, = get_lapack_funcs(('geqrf',), (other.u,))
-        qr, tau, work, info = geqrf(other.u, lwork = -1, overwrite_a = True) # sometimes segfaults with overwrite_a=True...?
-        qr, tau, work, info = geqrf(other.u, lwork = work[0], overwrite_a = True) # sometimes segfaults with overwrite_a=True...?
+        qr, tau, work, info = geqrf(other.u, lwork = -1, overwrite_a = True)
+        qr, tau, work, info = geqrf(other.u, lwork = work[0], overwrite_a = True)
         del other.u
         assert info >= 0
         r = triu(qr[:n2, :n2])
@@ -169,7 +217,7 @@ class Projection(utils.SaveLoad):
         assert info >= 0, "qr failed"
         assert q.flags.f_contiguous
         
-        # find rotation that diagonalizes r
+        # find the rotation that diagonalizes r
         k = numpy.bmat([[numpy.diag(decay * self.s), c * other.s], [matutils.pad(numpy.matrix([]).reshape(0, 0), min(m, n2), n1), r * other.s]])
         logger.debug("computing SVD of %s dense matrix" % str(k.shape))
         try:
@@ -184,7 +232,7 @@ class Projection(utils.SaveLoad):
             u_k, s_k, _ = numpy.linalg.svd(numpy.dot(k, k.T), full_matrices = False) # if this fails too, give up
             s_k = numpy.sqrt(s_k)
         
-        k = clipSpectrum(s_k, self.k)
+        k = clipSpectrum(s_k ** 2, self.k)
         u_k, s_k = u_k[:, :k], s_k[:k]
         
         # update & rotate current basis U
@@ -213,8 +261,8 @@ class LsiModel(interfaces.TransformationABC):
     Model persistency is achieved via its load/save methods.
     
     """
-    def __init__(self, corpus = None, id2word = None, numTopics = 200, 
-                 chunks = None, decay = 1.0, serial_only = None):
+    def __init__(self, corpus = None, numTopics = 200, id2word = None, chunks = 20000, 
+                 decay = 1.0, distributed = False, onepass = False):
         """
         `numTopics` is the number of requested factors (latent dimensions). 
         
@@ -227,27 +275,29 @@ class LsiModel(interfaces.TransformationABC):
         If you specify a `corpus`, it will be used to train the model. See the 
         method `addDocuments` for a description of the `chunks` and `decay` parameters.
         
-        The algorithm will automatically try to find active nodes on other computers
-        and run in a distributed manner; if this fails, it falls back to serial mode
-        (single core). To suppress distributed computing, set the `serial_only`
-        constructor parameter to True.
+        If your document stream is one-pass only (the stream cannot be repeated),
+        turn on `onepass` to force a single pass SVD algorithm (slower).
+
+        Turn on `distributed` to enforce distributed computing (only makes sense 
+        if `onepass` is set at the same time, too).
         
         Example:
         
-        >>> lsi = LsiModel(corpus, numTopics = 10)
-        >>> print lsi[doc_tfidf]
+        >>> lsi = LsiModel(corpus, numTopics=10)
+        >>> print lsi[doc_tfidf] # project some document into LSI space
         >>> lsi.addDocuments(corpus2) # update LSI on additional documents
         >>> print lsi[doc_tfidf]
         
         """
         self.id2word = id2word
-        self.numTopics = numTopics # number of latent topics
-        if chunks is None:
-            # by default, proceed in chunks as big as number of topics, to improve accuracy
-            self.chunks = max(numTopics, 100)
-        else:
-            self.chunks = chunks
-        self.decay = decay
+        self.numTopics = int(numTopics)
+        self.chunks = int(chunks)
+        self.decay = float(decay)
+        if distributed:
+            if not onepass:
+                logger.warning("forcing the one-pass algorithm for distributed LSA")
+                onepass = True
+        self.onepass = onepass
         
         if corpus is None and self.id2word is None:
             raise ValueError('at least one of corpus/id2word must be specified, to establish input space dimensionality')
@@ -261,49 +311,47 @@ class LsiModel(interfaces.TransformationABC):
         
         self.docs_processed = 0
         self.projection = Projection(self.numTerms, self.numTopics)
-
-        if serial_only:
-            logger.info("using slave LSI version on this node")
+        
+        if not distributed:
+            logger.info("using serial LSI version on this node")
             self.dispatcher = None
         else:
+            if not onepass:
+                raise NotImplementedError("distributed stochastic LSA not implemented yet; "
+                                          "run either distributed one-pass, or serial randomized.")
             try:
                 import Pyro
                 ns = Pyro.naming.locateNS()
                 dispatcher = Pyro.core.Proxy('PYRONAME:gensim.dispatcher@%s' % ns._pyroUri.location)
                 logger.debug("looking for dispatcher at %s" % str(dispatcher._pyroUri))
                 dispatcher.initialize(id2word = self.id2word, numTopics = numTopics, 
-                                      chunks = chunks, decay = decay, 
-                                      serial_only = True)
+                                      chunks = chunks, decay = decay,
+                                      distributed = False, onepass = onepass)
                 self.dispatcher = dispatcher
                 logger.info("using distributed version with %i workers" % len(dispatcher.getworkers()))
             except Exception, err:
-                if serial_only is not None: 
-                    # distributed version was specifically requested, so this is an error state
-                    logger.error("failed to initialize distributed LSI (%s)" % err)
-                    raise RuntimeError("failed to initialize distributed LSI (%s)" % err)
-                else:
-                    # user didn't request distributed specifically; just let him know we're running in serial
-                    logger.info("distributed LSI not available, running LSI in serial mode (%s)" % err)
-                self.dispatcher = None
+                # distributed version was specifically requested, so this is an error state
+                logger.error("failed to initialize distributed LSI (%s)" % err)
+                raise RuntimeError("failed to initialize distributed LSI (%s)" % err)
 
         if corpus is not None:
-            self.addDocuments(corpus, chunks = chunks)
+            self.addDocuments(corpus)
     
     
     def addDocuments(self, corpus, chunks = None, decay = None):
         """
-        Update singular value decomposition factors to take into account a new 
+        Update singular value decomposition to take into account a new 
         corpus of documents.
         
-        Training proceeds in chunks of `chunks` documents at a time. If the 
-        distributed mode is on, each chunk is sent to a different worker/computer.
-        Size of `chunks` is a tradeoff between increased speed (bigger `chunks`) vs. 
-        lower memory footprint (smaller `chunks`).
+        Training proceeds in chunks of `chunks` documents at a time. The size of 
+        `chunks` is a tradeoff between increased speed (bigger `chunks`) 
+        vs. lower memory footprint (smaller `chunks`). If the distributed mode 
+        is on, each chunk is sent to a different worker/computer.
 
         Setting `decay` < 1.0 causes re-orientation towards new data trends in the 
         input document stream, by giving less emphasis to old observations. This allows
-        SVD to gradually "forget" old observations and give more preference to 
-        new ones. The decay is applied once after every `chunks` documents.
+        SVD to gradually "forget" old observations (documents) and give more 
+        preference to new ones.
         """
         logger.info("updating SVD with new documents")
         
@@ -313,47 +361,52 @@ class LsiModel(interfaces.TransformationABC):
         if decay is None:
             decay = self.decay
         
-        if utils.isCorpus(corpus):
-            # do the actual work -- perform iterative singular value decomposition.
-            chunker = itertools.groupby(enumerate(corpus), key = lambda val: val[0] / chunks)
-            doc_no = 0
-            for chunk_no, (key, group) in enumerate(chunker):
-                # construct the job as a sparse matrix, to minimize memory overhead
-                # definitely avoid materializing it as a dense matrix!
-                job = matutils.corpus2csc(self.numTerms, (doc for _, doc in group))
-                doc_no += job.shape[1]
+        if not scipy.sparse.issparse(corpus):
+            if not self.onepass:
+                # we are allowed multiple passes over input => use a faster, randomized two-pass algo
+                update = Projection(self.numTerms, self.numTopics, corpus, algo = 'twopass', chunks = chunks)
+                self.projection.merge(update, decay = decay)
+            else:
+                # the one-pass algo
+                chunker = itertools.groupby(enumerate(corpus), key = lambda (docno, doc): docno / chunks)
+                doc_no = 0
+                for chunk_no, (key, group) in enumerate(chunker):
+                    # construct the job as a sparse matrix, to minimize memory overhead
+                    # definitely avoid materializing it as a dense matrix!
+                    job = matutils.corpus2csc((doc for _, doc in group), num_terms = self.numTerms)
+                    doc_no += job.shape[1]
+                    if self.dispatcher:
+                        # distributed version: add this job to the job queue, so workers can work on it
+                        logger.debug("creating job #%i" % chunk_no)
+                        self.dispatcher.putjob(job) # put job into queue; this will eventually block, because the queue has a small finite size
+                        del job
+                        logger.info("dispatched documents up to #%s" % doc_no)
+                    else:
+                        # serial version, there is only one "worker" (myself) => process the job directly
+                        update = Projection(self.numTerms, self.numTopics, job)
+                        del job
+                        self.projection.merge(update, decay = decay)
+                        del update
+                        logger.info("processed documents up to #%s" % doc_no)
+                        self.printTopics(5) # TODO see if printDebug works and remove one of these..
+                
                 if self.dispatcher:
-                    # distributed version: add this job to the job queue, so workers can work on it
-                    logger.debug("creating job #%i" % chunk_no)
-                    self.dispatcher.putjob(job) # put job into queue; this will eventually block, because the queue has a small finite size
-                    del job
-                    logger.info("dispatched documents up to #%s" % doc_no)
-                else:
-                    # serial version, there is only one "worker" (myself) => process the job directly
-                    update = Projection(self.numTerms, self.numTopics, job)
-                    del job
-                    self.projection.merge(update, decay = decay)
-                    del update
-                    logger.info("processed documents up to #%s" % doc_no)
-                    #self.printDebug(5)
-                    self.printTopics(5) # TODO see if printDebug works and remove one of these..
-            
-            if self.dispatcher:
-                logger.info("reached the end of input; now waiting for all remaining jobs to finish")
-                import time
-                while self.dispatcher.jobsdone() <= chunk_no:
-                    time.sleep(0.5) # check every half a second
-                logger.info("all jobs finished, downloading final projection")
-                del self.projection
-                self.projection = self.dispatcher.getstate()
-                logger.info("decomposition complete")
+                    logger.info("reached the end of input; now waiting for all remaining jobs to finish")
+                    import time
+                    while self.dispatcher.jobsdone() <= chunk_no:
+                        time.sleep(0.5) # check every half a second
+                    logger.info("all jobs finished, downloading final projection")
+                    del self.projection
+                    self.projection = self.dispatcher.getstate()
+                    logger.info("decomposition complete")
         else:
             assert not self.dispatcher, "must be in serial mode to receive jobs"
-            assert isinstance(corpus, scipy.sparse.csc_matrix)
-            update = Projection(self.numTerms, self.numTopics, corpus)
+            assert self.onepass, "distributed two-pass algo not supported yet"
+            update = Projection(self.numTerms, self.numTopics, corpus.tocsc())
             self.projection.merge(update, decay = decay)
             logger.info("processed sparse job of %i documents" % (corpus.shape[1]))
-            self.printTopics(5)
+
+        self.printTopics(5)
 
     
     def __str__(self):
@@ -371,8 +424,9 @@ class LsiModel(interfaces.TransformationABC):
         singular values**. To return non-scaled embedding, set `scaled` to False.
         """
         # if the input vector is in fact a corpus, return a transformed corpus as result
-        if utils.isCorpus(bow):
-            return self._apply(bow)
+        is_corpus, bow = utils.isCorpus(bow)
+        if is_corpus:
+            self._apply(bow)
         
         assert self.projection.u is not None, "decomposition not initialized yet"
         vec = numpy.asfortranarray(matutils.sparse2full(bow, self.numTerms), dtype = self.projection.u.dtype)
@@ -477,16 +531,19 @@ def svdUpdate(U, S, V, a, b):
     `[X + a * b^T] = U' * S' * V'^T`
     and return `U'`, `S'`, `V'`.
     
-    The original matrix X is not needed at all, so this function implements flexible 
-    *online* updates to an existing decomposition. 
+    The original matrix X is not needed at all, so this function implements one-pass
+    streaming rank-1 updates to an existing decomposition. 
     
     `a` and `b` are (m, 1) and (n, 1) matrices.
     
     You can set V to None if you're not interested in the right singular
     vectors. In that case, the returned V' will also be None (saves memory).
     
+    The blocked merge algorithm in LsiModel.addDocuments() is much faster; I keep this fnc here
+    purely for backup reasons.
+
     This is the rank-1 update as described in
-    *Brand, 2006: Fast low-rank modifications of the thin singular value decomposition*,
+    **Brand, 2006: Fast low-rank modifications of the thin singular value decomposition**,
     but without separating the basis from rotations.
     """
     # convert input to matrices (no copies of data made if already numpy.ndarray or numpy.matrix)
@@ -553,14 +610,13 @@ def iterSvd(corpus, numTerms, numFactors, numIter = 200, initRate = None, conver
     
     The algorithm performs at most `numFactors*numIters` passes over the corpus.
     
-    See **Genevieve Gorrell: Generalized Hebbian Algorithm for Incremental Singular 
-    Value Decomposition in Natural Language Processing. EACL 2006.**
-    
     Use of this function is deprecated; although it works, it is several orders of 
     magnitude slower than our own, direct (non-stochastic) version (which
-    operates in a single pass, too, and can be distributed). 
+    operates in a single pass, too, and can be distributed). I keep this function 
+    here purely for backup reasons.
     
-    I keep this function here purely for backup reasons.
+    See **Genevieve Gorrell: Generalized Hebbian Algorithm for Incremental Singular 
+    Value Decomposition in Natural Language Processing. EACL 2006.**
     """
     logger.info("performing incremental SVD for %i factors" % numFactors)
 
@@ -631,25 +687,86 @@ def iterSvd(corpus, numTerms, numFactors, numIter = 200, initRate = None, conver
     return sterm, svals, sdoc.T
 
 
-#def stochasticSvd(a, numTerms, numFactors, p = None, q = 0):
-#    """
-#    SVD decomposition based on stochastic approximation.
-#    
-#    See **Halko, Martinsson, Tropp. Finding structure with randomness, 2009.**
-#    
-#    This is the randomizing version with oversampling, but without power iteration. 
-#    """
-#    k = numFactors
-#    if p is None:
-#        l = 2 * k # default oversampling
-#    else:
-#        l = k + p
-#    
-#    # stage A: construct the "action" basis matrix Q
-#    y = numpy.empty(dtype = numpy.float64, shape = (numTerms, l)) # in double precision, because we will be computing orthonormal basis on this possibly ill-conditioned projection
-#    for i, row in enumerate(a):
-#        y[i] = column_stack(matutils.sparse2full(doc, numTerms) * numpy.random.normal(0.0, 1.0, numTerms)
-#                           for doc in corpus)
-#    q = numpy.linalg.qr()
-#    
+def stochasticSvd(corpus, rank, num_terms=None, chunks=20000, extra_dims=None, dtype=numpy.float64, eps=1e-6):
+    """
+    Return U, S -- the left singular vectors and the singular values of the streamed 
+    input corpus.
+    
+    This may actually return less than the requested number of top `rank` factors, 
+    in case the input is of lower rank. Also note that the decomposition is unique
+    up the the sign of the left singular vectors (columns of U).
+    
+    This is a streamed, two-pass algorithm, without power-itrations. In case you can 
+    only afford a single pass over the input corpus, set onepass=True in LsiModel and 
+    avoid using this algorithm.
+
+    The decomposition algorithm is based on 
+    **Halko, Martinsson, Tropp. Finding structure with randomness, 2009.**
+    
+    """
+    rank = int(rank)
+    if extra_dims is None:
+        samples = max(10, 2 * rank) # use more samples than requested factors, to improve accuracy
+    else:
+        samples = rank + int(extra_dims)
+    
+    if num_terms is None:
+        logger.warning("number of terms not provided; will scan the corpus (ONE EXTRA PASS, MAY BE SLOW) to determine it")
+        num_terms = len(utils.dictFromCorpus(corpus))
+        logger.info("found %i terms" % num_terms)
+    else:
+        num_terms = int(num_terms)
+    
+    eps = max(float(eps), 1e-9) # must ignore near-zero eigenvalues (probably numerical error); the associated eigenvectors are typically unstable/garbage
+    
+    # first pass: construct the orthonormal action matrix Q = orth(Y) = orth(A * O)
+    # build Y in blocks of `chunks` documents (much faster than going one-by-one 
+    # and more memory friendly than processing all documents at once)
+    y = numpy.zeros(dtype = dtype, shape = (num_terms, samples))
+    logger.info("1st pass: constructing %s action matrix" % str(y.shape))
+    
+    chunker = itertools.groupby(enumerate(corpus), key = lambda (docno, doc): docno / chunks)
+    for chunk_no, (key, group) in enumerate(chunker):
+        logger.info('PROGRESS: at document #%i' % (chunk_no * chunks))
+        # construct the chunk as a sparse matrix, to minimize memory overhead
+        # definitely avoid materializing it as a dense (num_terms x chunks) matrix!
+        chunk = matutils.corpus2csc((doc for _, doc in group), num_terms=num_terms, dtype=dtype) # documents = columns of sparse CSC
+        m, n = chunk.shape
+        assert m == num_terms
+        assert n <= chunks # the very last chunk of A may be smaller
+        logger.debug("multiplying chunk * gauss")
+        o = numpy.random.normal(0.0, 1.0, (n, samples)).astype(dtype) # draw a random gaussian matrix
+        sparsetools.csc_matvecs(num_terms, n, samples, chunk.indptr, # y = y + chunk * o
+                                chunk.indices, chunk.data, o.ravel(), y.ravel())
+        del chunk, o
+    
+    logger.info("orthonormalizing %s action matrix" % str(y.shape))
+    q, r = numpy.linalg.qr(y) # orthonormalize the range
+    del y # Y not needed anymore, free up mem
+    samples = clipSpectrum(numpy.diag(r), samples, discard = eps)
+    qt = q[:, :samples].T.copy() # discard bogus columns, in case Y was rank-deficient
+    del q
+    
+    # second pass: construct the covariance matrix X = B * B.T, where B = Q.T * A
+    # again, construct X incrementally, in chunks of `chunks` documents from the streaming 
+    # input corpus A, to avoid using O(number of documents) memory
+    x = numpy.zeros(shape = (samples, samples), dtype = dtype)
+    logger.info("2nd pass: constructing %s covariance matrix" % str(x.shape))
+    chunker = itertools.groupby(enumerate(corpus), key = lambda (docno, doc): docno / chunks)
+    for chunk_no, (key, group) in enumerate(chunker):
+        logger.info('PROGRESS: at document #%i' % (chunk_no * chunks))
+        chunk = matutils.corpus2csc((doc for _, doc in group), num_terms=num_terms, dtype=dtype)
+        b = qt * chunk # dense * sparse matrix multiply
+        x += numpy.dot(b, b.T) # TODO should call the BLAS routine SYRK, but there is no SYRK wrapper in scipy :(
+        del chunk, b
+    
+    # now we're ready to compute decomposition of the small matrix X
+    logger.info("computing decomposition of the %s covariance matrix" % str(x.shape))
+    u, s, vt = numpy.linalg.svd(x) # could use linalg.eigh, but who cares... and svd returns the factors already sorted :)
+    keep = clipSpectrum(s, rank, discard = eps)
+    
+    logger.info("computing the final decomposition")
+    s = numpy.sqrt(s[:keep]) # sqrt to go back from singular values of X to singular values of B = singular values of the corpus
+    u = numpy.dot(qt.T, u[:, :keep]) # go back from left singular vectors of B to left singular vectors of the corpus
+    return u.astype(dtype), s.astype(dtype)
     
